from setuptools import setup
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='thermobuilpy',
    version='1.0.1',
    description='This package provides a model for the formal description of thermal RC systems and enables their simulation using various discretization methods.',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Hannes Hanse',
    author_email='hannes.hanse@tu-clausthal.de',
    keywords=['Thermal Model','RC Model','Energy System','Stratified Storage'],
    readme = "README.md",
    url='https://github.com/hanneshanse/ThermoBuilPy',
    install_requires=["numpy","matplotlib"],
    classifiers = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
]
)